/*
 *  USB device using FTDI's FT245BM USB FIFO chip.
 */

#define USE_FTDILIB 1

typedef unsigned char u08;
typedef unsigned short u16;
typedef unsigned long u32;
typedef long  s32;

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#define FTDI_SPEED 3000000

#include <ftdi.h>
#ifdef USE_FTDILIB

int           verbose = 1;

static struct ftdi_context ctx;

static int
Error( const char *where, int rv, struct ftdi_context *ctx, int doclose )
{
  if ( rv )
  {
    fprintf( stderr, "%s: rv=%d, error=%s, errno=%s (%d)\n",
	where, rv, ctx->error_str, strerror( errno ), errno );
    if ( doclose )
    {
      rv = ftdi_usb_close( ctx );
      fprintf( stderr, "close: rv=%d, error=%s, errno=%s (%d)\n",
	  rv, ctx->error_str, strerror( errno ), errno );
    }
    ftdi_deinit( ctx );
    return 1;
  }
  if ( verbose )
    fprintf( stderr, "%s: OK\n", where );
  return 0;
}

int
get_byte( void )
{
  int           rv;
  u08           byte;

  while ( ( rv = ftdi_read_data( &ctx, &byte, 1 ) ) == 0 ) ;
  if ( rv < 0 )
  {
    Error( "read", rv, &ctx, 1 );
    return -1;
  }
  return byte;
}

int
get_word( void )
{
  int           rv;
  u16           word;

  while ( ( rv = ftdi_read_data( &ctx, ( void * ) &word, 2 ) ) == 0 ) ;
  if ( rv < 0 )
  {
    Error( "read", rv, &ctx, 1 );
    return -1;
  }
  return word;
}

u32
get_dword( void )
{
  int           rv;
  u32           dword;

  while ( ( rv = ftdi_read_data( &ctx, ( void * ) &dword, 4 ) ) == 0 ) ;
  if ( rv < 0 )
  {
    Error( "read", rv, &ctx, 1 );
    return -1;
  }
  return dword;
}

int
put_byte( u08 byte )
{
  int           rv;

//  ftdi_write_data_set_chunksize( &ctx, 1 );
  rv = ftdi_write_data( &ctx, &byte, 1 );
  if ( rv <= 0 )
  {
    Error( "write", rv, &ctx, 1 );
    return 0;
  }
  return rv;
}

int
put_dword( u32 dword )
{
  int           rv;

//  ftdi_write_data_set_chunksize( &ctx, 4 );
  rv = ftdi_write_data( &ctx, ( void * ) &dword, 4 );
  if ( rv <= 0 )
  {
    Error( "write", rv, &ctx, 1 );
    return 0;
  }
  return rv;
}

void
get_data( u08 * buf, int len )
{
  int           rv;

  do
  {
    rv = ftdi_read_data( &ctx, buf, len );
    if ( rv < 0 )
      Error( "read", rv, &ctx, 1 );
  }
  while ( rv == 0 );
}

int
put_data( u08 * buf, int len )
{
  int           rv;

//  ftdi_write_data_set_chunksize( &ctx, len );
  rv = ftdi_write_data( &ctx, buf, len );
  if ( rv <= 0 )
  {
    Error( "write", rv, &ctx, 1 );
    return 0;
  }
  return rv;
}

void
flush_ftdi_buffers( void )
{
  u08           flush[1024];

  ftdi_read_data( &ctx, flush, 1024 );	// clear receive buffer
}

int
open_ftdi( void )
{
  int           rv;
  unsigned int  rdchk, wrchk;

  rv = ftdi_init( &ctx );
  if ( Error( "init", rv, &ctx, 0 ) )
    return 0;

  rv = ftdi_usb_open( &ctx, 0x0403, 0x6001 );	// vendor, device
  if ( Error( "open", rv, &ctx, 0 ) )
    return 0;

  rv = ftdi_usb_reset( &ctx );
  if ( Error( "reset", rv, &ctx, 1 ) )
    return 0;

  rv = ftdi_disable_bitbang( &ctx );
  if ( Error( "bitbang(off)", rv, &ctx, 1 ) )
    return 0;

  rv = ftdi_set_baudrate( &ctx, FTDI_SPEED );
  if ( Error( "baudrate", rv, &ctx, 1 ) )
    return 0;

  ftdi_write_data_set_chunksize( &ctx, 4096 );
  ftdi_read_data_set_chunksize( &ctx, 4096 );

  rv = ftdi_read_data_get_chunksize( &ctx, &rdchk );
  rv = ftdi_write_data_get_chunksize( &ctx, &wrchk );
  if ( verbose )
    fprintf( stderr, "Chunk sizes RD %d, WR %d\n", rdchk, wrchk );

  return 1;
}

void
close_ftdi( void )
{
  int           rv;

  rv = ftdi_usb_close( &ctx );
  Error( "close", rv, &ctx, 0 );
  ftdi_deinit( &ctx );
}

#else				// USE_FTDILIB (use the FTDI kernel driver)

#include <termios.h>
#include <unistd.h>
#include <string.h>
#include <linux/serial.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#define SERIAL_DEVICE "/dev/ttyUSB0"
#define BAUD_RATE B38400

static int    fds;
static struct termios oldtio;

int
get_byte( void )
{
  u08           byte;

  while ( read( fds, &byte, 1 ) == 0 )
    ;
  return byte;
}

int
get_word( void )
{
  u16           word;

  while ( read( fds, &word, 2 ) == 0 )
    ;
  return word;
}

u32
get_dword( void )
{
  u32           dword;

  while ( read( fds, &dword, 4 ) == 0 )
    ;
  return dword;
}

int
put_byte( u08 byte )
{
  return write( fds, &byte, 1 );
}

int
put_dword( u32 dword )
{
  return write( fds, &dword, 4 );
}

void
get_data( u08 * buf, int len )
{
  while ( read( fds, buf, len ) == 0 )
    ;
}

int
put_data( u08 * buf, int len )
{
  return write( fds, buf, len );
}

void
flush_ftdi_buffers( void )
{
  u08           flush[1024];

  while ( read( fds, flush, 1024 ) != 0 )	// clear receive buffer
    ;
}

int
open_ftdi( void )
{
  struct termios newtio;
  struct serial_struct warp;

  fds = open( SERIAL_DEVICE, O_RDWR | O_NOCTTY );
  if ( fds < 0 )
  {
    perror( SERIAL_DEVICE );
    exit( -1 );
//    return  -1;
  }

  tcgetattr( fds, &oldtio );	// save current port settings 

  bzero( &newtio, sizeof( newtio ) );
  newtio.c_cflag = BAUD_RATE | CS8 | CLOCAL | CREAD;
  newtio.c_iflag = IGNPAR;
  newtio.c_oflag = 0;
  // set input mode (non-canonical, no echo,...) 
  newtio.c_lflag = 0;

  newtio.c_cc[VTIME] = 1;	// read waits for 100ms before returning
  newtio.c_cc[VMIN] = 0;	// non-blocking read

  tcflush( fds, TCIFLUSH );
  tcsetattr( fds, TCSANOW, &newtio );

  ioctl( fds, TIOCGSERIAL, &warp );
  warp.flags &= ~ASYNC_SPD_MASK;
  warp.flags |= ASYNC_SPD_CUST;
  warp.custom_divisor = warp.baud_base / FTDI_SPEED;
  ioctl( fds, TIOCSSERIAL, &warp );
  return 1;
}

void
close_ftdi( void )
{
  tcsetattr( fds, TCSANOW, &oldtio );
}

#endif				// USE_FTDILIB


FILE *l;

typedef void (*sighandler_t)(int);

void
s_break( int a )
{
  close_ftdi(  );
  fclose(l);
  exit(0);
}

int
main(  )
{
  char buf[1024];		// wrap arround
  int pb, para, cmd;
  unsigned char c;
  
  open_ftdi(  );
  flush_ftdi_buffers();
  signal(SIGINT, s_break );
  while( 1 ) 
  {
    c = get_byte();
    putchar(c ) ;
//    printf( "%s\n", buf );
//    fprintf( l, "%s\n", buf );
  }
  return 0;
}
